@@ -1,27 +1,11 @@
 # High-level parallelization classes
 #
-# Written by Konrad Hinsen <hinsen@cnrs-orleans.fr>
-# last revision: 2003-3-20
+# Written by Konrad Hinsen
 #
 
-"""This module contains high-level parallelization constructs
-based on the Bulk Synchronous Parallel (BSP) model.
-
-Parallelization requires a low-level communications library, which can
-be either BSPlib or MPI. Programs must be run with the bsppython or
-mpipython executables in order to use several processors. When run
-with a standard Python interpreter, only one processor is available.
-
-A warning about object identity: when a communication operation
-transmits a Python object to the same processor, the object in the
-return list can either be the sent object or a copy of it. Application
-programs thus should not make any assumptions about received objects
-being different from sent objects.
-"""
-
-_undocumented = 1
-
-import cPickle, Numeric, operator, sys, types
+import RemoteObjects
+from Scientific import N
+import cPickle, operator, sys, types
 
 try:
     virtual_bsp_machine = sys.virtual_bsp_machine
@@ -30,8 +14,7 @@ try:
 except AttributeError:
     virtual_bsp_machine = None
     try:
-        import Scientific_bsplib
-        bsplib = Scientific_bsplib
+        from Scientific import _bsplib as bsplib
         world = None
     except ImportError:
         bsplib = None
@@ -65,10 +48,10 @@ else:
 #
 if world is not None:
 
-    _type_tags = {Numeric.Int8: 3, Numeric.Int16: 4, Numeric.Int32: 5,
-                  Numeric.UnsignedInt8: 6,
-                  Numeric.Float16: 7, Numeric.Float: 8,
-                  Numeric.Complex32: 9, Numeric.Complex64: 10}
+    _type_tags = {N.Int8: 3, N.Int16: 4, N.Int32: 5,
+                  N.UnsignedInt8: 6,
+                  N.Float16: 7, N.Float: 8,
+                  N.Complex32: 9, N.Complex64: 10}
     _type_tags_i = {}
     for key, value in _type_tags.items():
         _type_tags_i[value] = key
@@ -78,20 +61,23 @@ if world is not None:
         global _debug_flag
         _debug_flag = flag
 
-    def _send(object, destinations):
+    def _send(obj, destinations):
         requests = []
-        if type(object) is Numeric.arraytype:
-            send_data = object
-            tag = _type_tags.get(send_data.typecode(), 2)
+        if type(obj) is N.arraytype:
+            send_data = obj
+            try:
+                type_code = send_data.typecode() # Numeric, numarray
+            except AttributeError:
+                type_code = send_data.dtype.char # NumPy
+            tag = _type_tags.get(type_code, 2)
             if _debug_flag:
                 print world.rank, "sending array (type %s, shape %s) to %s" \
-                      % (send_data.typecode(), str(object.shape), \
-                         str(destinations))
+                      % (type_code, str(obj.shape), str(destinations))
                 sys.stdout.flush()
             if tag == 2:
                 send_data = cPickle.dumps(send_data, 1)
             else:
-                shape = Numeric.array(object.shape)
+                shape = N.array(obj.shape)
                 for pid in destinations:
                     requests.append(world.nonblockingSend(shape, pid, tag))
                 tag = 1
@@ -99,7 +85,7 @@ if world is not None:
             if _debug_flag:
                 print world.rank, "sending non-array object to", destinations
                 sys.stdout.flush()
-            send_data = cPickle.dumps(object, 1)
+            send_data = cPickle.dumps(obj, 1)
             tag = 2
         if _debug_flag:
             print world.rank, "sending data (%d) to" % tag, destinations
@@ -131,8 +117,8 @@ if world is not None:
                 sys.stdout.flush()
             typecode = _type_tags_i.get(tag, None)
             if typecode is None:
-                raise ValueError, "Invalid tag " + `tag`
-            shape = world.receive(Numeric.Int, source, tag)[0]
+                raise ValueError("Invalid tag " + `tag`)
+            shape = world.receive(N.Int, source, tag)[0]
             if _debug_flag:
                 print world.rank, "shape: ", shape
                 print world.rank, "receiving array data from", source
@@ -155,11 +141,11 @@ if virtual_bsp_machine is not None:
 
 elif bsplib is not None:
 
-    def put(object, pid_list):
-        if type(object) is not Numeric.arraytype:
-            object = cPickle.dumps(object, 1)
+    def put(obj, pid_list):
+        if type(obj) is not N.arraytype:
+            obj = cPickle.dumps(obj, 1)
         for pid in pid_list:
-            bsplib.send(object, pid)
+            bsplib.send(obj, pid)
             
     def send(messages):
         for pid, data in messages:
@@ -171,7 +157,7 @@ elif bsplib is not None:
         while 1:
             data = bsplib.receive()
             if data is None: break
-            if type(data) is Numeric.arraytype:
+            if type(data) is N.arraytype:
                 messages.append(data)
             else:
                 messages.append(cPickle.loads(data))
@@ -181,10 +167,10 @@ elif world is not None:
 
     _requests = []
 
-    def put(object, pid_list):
+    def put(obj, pid_list):
         global _requests
         if len(pid_list) > 0:
-            _requests = _requests + _send(object, pid_list)
+            _requests = _requests + _send(obj, pid_list)
 
     def send(messages):
         global _requests
@@ -215,13 +201,13 @@ else:
 
     _messages = []
 
-    def put(object, pid_list):
+    def put(obj, pid_list):
         global _messages
         for pid in pid_list:
             if pid == 0:
-                _messages.append(object)
+                _messages.append(obj)
             else:
-                raise ValueError, "invalid pid"
+                raise ValueError("invalid pid")
 
     def send(messages):
         for pid, data in messages:
@@ -233,6 +219,20 @@ else:
         _messages = []
         return messages
 
+#
+# Higher-level communications layer. This code takes care of the handling
+# of special objects and of special transfer needs of particular objects.
+#
+
+def retrieveMessages():
+    messages = sync()
+    filtered_messages = []
+    for m in messages:
+        if isinstance(m, RemoteObjects.TransferToken):
+            RemoteObjects.remote_object_manager.handleTransfer(m)
+        else:
+            filtered_messages.append(m)
+    return filtered_messages
 
 #
 # The dictionary _wrappers stores the global class corresponding
@@ -253,9 +253,10 @@ def global_object(local_object):
 #
 # ParValue is the base class for all standard distributed-data classes.
 #
-class ParValue:
+class ParValue(object):
 
-    """Global data
+    """
+    Global data
 
     ParValue instances are created internally, but are not meant to be
     created directly by application programs. Use the subclasses instead.
@@ -267,7 +268,14 @@ class ParValue:
     if their local values are callable.
     """
 
-    def __init__(self, value, valid=1):
+    def __init__(self, value, valid=True):
+        """
+        @param value: the local value
+        @type value: any
+        @param valid: C{True} if the value is valid, C{False} if it is not.
+        Invalid values are not treated, any operation on them produces
+        an invalid result.
+        """
         self.value = value
         self.valid = valid
 
@@ -285,20 +293,15 @@ class ParValue:
     __str__ = __repr__
 
     def __call__(self, *args, **kwargs):
-        def getValue(x):
-            if hasattr(x, 'is_parvalue'):
-                return x.value, x.valid
-            else:
-                return x, 1
         params = []
         valid = self.valid
         for a in args:
-            p, v = getValue(a)
+            p, v = _getValue(a)
             valid = valid and v
             params.append(p)
         kw = {}
         for key, data in kwargs.items():
-            p, v = getValue(data)
+            p, v = _getValue(data)
             kw[key] = p
             valid = valid and v
         if valid:
@@ -307,66 +310,95 @@ class ParValue:
             return ParValue(None, 0)
 
     def put(self, pid_list):
-        """Sends the local data to all processors in |pid_list| (a global
-        object). Returns a ParValue object whose local value is a list of
+        """
+        Send the local data to all specified processors.
+
+        @param pid_list: a list of processor IDs to which the data is sent.
+        @type pid_list: Global C{list} of C{int}
+        @returns: a global object whose local value is a list of
         all the data received from other processors. The order of the
         data in that list is not defined.
+        @rtype: L{ParValue}
         """
         if self.valid:
             if not pid_list.valid:
-                raise ValueError, "Invalid processor ID list"
+                raise ValueError("Invalid processor ID list")
             put(self.value, pid_list.value)
-        return ParValue(sync())
+        return ParValue(retrieveMessages())
 
     def get(self, pid_list):
-        """Requests the local data from all processors in |pid_list| (a global
-        object). Returns a ParValue object whose local value is a list of
+        """
+        Request the local data from all specified processors.
+
+        @param pid_list: a list of processor IDs to which the data is sent.
+        @type pid_list: Global C{list} of C{int}
+        @returns: a global object whose local value is a list of
         all the data received from other processors. The order of the
         data in that list is not defined.
+        @rtype: L{ParValue}
         """
         if not pid_list.valid:
-            raise ValueError, "Invalid processor ID list"
+            raise ValueError("Invalid processor ID list")
         put(processorID, pid_list.value)
         destinations = sync()
         if self.valid:
             put(self.value, destinations)
-        return ParValue(sync())
+        return ParValue(retrieveMessages())
 
     def broadcast(self, from_pid=0):
-        """Transmits the local data on processor |from_pid| to all
-        processors. Returns a ParValue object.
+        """
+        Transmit the local data of one processor to all processors.
+
+        @param from_pid: the ID of the processor that sends data
+        @type from_pid: C{int}
+        @returns: the transmitted data on all processors
+        @rtype: L{ParValue}
         """
         if processorID == from_pid:
             if self.valid:
                 put(self.value, range(numberOfProcessors))
             else:
-                raise ValueError, "Broadcast for invalid data"
-        return ParValue(sync()[0])
+                raise ValueError("Broadcast for invalid data")
+        return ParValue(retrieveMessages()[0])
 
     def fullExchange(self):
-        """Transmits the local data of each processor to all other
-        processors. Returns a ParValue object.
+        """
+        Transmit the local data of each processor to all other processors.
+
+        @returns: the transmitted data on all processors
+        @rtype: L{ParValue}
         """
         if self.valid:
             put(self.value, range(numberOfProcessors))
-        return ParValue(sync())
+        return ParValue(retrieveMessages())
 
     def reduce(self, operator, zero):
-        """Performs a reduction with |operator| over the local values
-        of all processors using |zero| as initial value. The result
-        is a ParValue object with the reduction result on processor 0
-        and |zero| on all other processors.
+        """
+        Perform a reduction over the local values of all processors.
+
+        @param operator: the binary operator used in the reduction
+        @type operator: function of two variables
+        @param zero: the initial value of the reduction
+        @type zero: any
+        @returns: the reduction on processor 0, zero elsewhere
+        @rtype: L{ParValue}
         """
         if self.valid:
             put(self.value, [0])
-        return ParValue(reduce(operator, sync(), zero), processorID == 0)
+        return ParValue(reduce(operator, retrieveMessages(), zero),
+                        processorID == 0)
 
     def accumulate(self, operator, zero):
-        """Performs an accumulation with |operator| over the local values
-        of all processors using |zero| as initial value. The result
-        is a ParValue object whose local value on each processor is the
-        reduction of the values from all processors with lower or equal
-        number.
+        """
+        Perform an accumulation over the local values of all processors.
+
+        @param operator: the binary operator used in the accumulation
+        @type operator: function of two variables
+        @param zero: the initial value of the accumulation
+        @type zero: any
+        @returns: on each processor, the reduction of the values from
+        processors with a lower or equal number
+        @rtype: L{ParValue}
         """
         if self.valid:
             data = self
@@ -377,27 +409,34 @@ class ParValue:
 
     def __nonzero__(self):
         if not self.valid:
-            raise ValueError, "invalid local value"
+            raise ValueError("invalid local value")
         return operator.truth(self.value)
 
     def alltrue(self):
-        """Returns 1 (local value) if the local values on all
-        processors are true.
+        """
+        @returns: C{True} if the local values on all processors are true.
+        @rtype: Local C{bool}
         """
         if self.valid:
             put(operator.truth(self.value), [0])
-        combined = reduce(operator.and_, sync(), 1)
-        put(combined, range(numberOfProcessors))
+        all = sync()
+        if processorID == 0:
+            combined = reduce(operator.and_, all, 1)
+            put(combined, range(numberOfProcessors))
         return sync()[0]
 
     def anytrue(self):
-        """Returns 1 (local value) if at least one of the local values on all
+        """
+        @returns: C{True} if at least one of the local values on all
         processors is true.
+        @rtype: Local C{bool}
         """
         if self.valid:
             put(operator.truth(self.value), [0])
-        combined = reduce(operator.or_, sync(), 0)
-        put(combined, range(numberOfProcessors))
+        all = sync()
+        if processorID == 0:
+            combined = reduce(operator.or_, all, 0)
+            put(combined, range(numberOfProcessors))
         return sync()[0]
 
     def __eq__(self, other):
@@ -466,6 +505,8 @@ class ParValue:
         else:
             return ParValue(None, 0)
 
+    __truediv__ = __div__
+
     def __mod__(self, other):
         if self.valid and other.valid:
             return ParValue(self.value % other.value)
@@ -509,7 +550,7 @@ class ParValue:
             else:
                 return ParValue(None, 0)
         else:
-            return global_object(self.value[item])
+            return global_object(_getValue(self.value[item])[0])
 
     def __getattr__(self, attr):
         if attr == 'valid' or attr == '__coerce__':
@@ -530,6 +571,14 @@ class ParValue:
             function = function.value
         return ParValue(map(function, self.value))
 
+#
+# Extract local value and validity flag from a ParValue
+def _getValue(x):
+    if isinstance(x, ParValue):
+        return x.value, x.valid
+    else:
+        return x, 1
+
 #
 # ParConstant represents an identical value on each processor.
 #
@@ -537,42 +586,77 @@ class ParConstant(ParValue):
 
     """Global constant
 
-    A subclass of ParValue.
-
-    Constructor: ParConstant(|value|)
-
-    Arguments:
-
-    |value| -- any local or global object
+    A subclass of ParValue that stores an identical value on each processor.
+    It must be called with the same argument on all processors, but this
+    is not verified in the current implementation.
     """
 
     def __init__(self, value):
+        """
+        @param value: any local or global object
+        @type value: any
+        """
         if hasattr(value, 'is_parvalue'):
             self.value = value.value
         else:
             self.value = value
         self.valid = 1
 
+#
+# This dummy class is used on nonroot machines because the value on
+# the root machine might be an object on which instance methods are
+# called.
+#
+class RootNoneObject(object):
+    def __getattr__(self, attr):
+        return None
+    
+    def __getitem__(self, item):
+        return None
+
+#
+# ParRootConstant is a constant on the root machine and None on all
+# other machines.
+#
+class ParRootConstant(ParConstant):
+    """
+    Global constant present only on processor 0.
+    """
+    def __init__(self, value):
+        """
+        @param value: the value to store.
+        @type value: any
+        """
+        self.value = RootNoneObject()
+        if processorID == 0:
+            if hasattr(value, 'is_parvalue'):
+                self.value = value.value
+            else:
+                self.value = value
+        self.valid = 1
+        
+
 #
 # ParData generates the local values as a function of processor id
 # and number of processors.
 #
 class ParData(ParValue):
 
-    """Global data
+    """
+    Global data
 
-    A subclass of ParValue
-
-    Constructor: ParData(|function|)
-
-    Arguments:
-
-    |function| -- a function of two arguments (processor number and
-                  number of processors in the machine) whose return
-                  value becomes the local value of the global object.
+    A subclass of ParValue that calculates its local value from the
+    processor number.
     """
 
     def __init__(self, function):
+        """
+        @param function: a function that is called with two arguments
+                         (processor number and number of processors
+                         in the system) and whose return value becomes
+                         the local value of the global object.
+        @type function: function of two arguments
+        """
         self.value = function(processorID, numberOfProcessors)
         self.valid = 1
 
@@ -581,22 +665,21 @@ class ParData(ParValue):
 #
 class ParSequence(ParValue):
 
-    """Global distributed sequence
+    """
+    Global distributed sequence
 
-    A subclass of ParValue.
-
-    Constructor: ParSequence(|full_sequence|)
-
-    Arguments:
-
-    |full_sequence| -- any indexable and sliceable Python sequence
-
-    The local value of a ParSequence object is a slice of |full_sequence|,
+    The local value of a ParSequence object is a slice of the input sequence,
     which is constructed such that the concatenation of the local values
-    of all processors equals |full_sequence|.
+    of all processors equals the input sequence while making the number of
+    elements on each processor as equal as possible.
     """
 
     def __init__(self, full_sequence):
+        """
+        @param full_sequence: the full sequence, equal to the concatenation
+        of the local values of all processors
+        @type full_sequence: arbitrary sequence object
+        """
         if hasattr(full_sequence, 'is_parvalue'):
             if not full_sequence.valid:
                 self.valid = 0
@@ -611,9 +694,21 @@ class ParSequence(ParValue):
         self.valid = 1
 
     def totalLength(self):
+        """
+        @returns: the sum of the lengths of the local values
+        @rtype: C{int}
+        """
         return ParValue(self.length)
 
     def __getitem__(self, item):
+        """
+        @param item: an index into the total sequence
+        @type item: C{int} or L{ParIndex}
+        @returns: the element referred to by the index, if it is in the local
+        subset
+        @rtype: any
+        @raise IndexError: if the index refers to an item on another processor
+        """
         if not self.valid:
             return ParValue(None, 0)
         if hasattr(item, 'is_parindex'):
@@ -637,63 +732,97 @@ class ParSequence(ParValue):
         else:
             return global_object(self.value[item-self.first])
 
+#
+# ParRootSequence objects distribute a sequence stored initially on
+# processor 0 over the processors
+#
+class ParRootSequence(ParSequence):
+
+    """
+    Global distributed sequence with data taken from processor 0
+
+    The local value of a ParRootSequence object is a slice of the input
+    sequence, which is constructed such that the concatenation of the
+    local values of all processors equals the input sequence while making
+    the number of elements on each processor as equal as possible.
+    """
+
+    def __init__(self, full_sequence):
+        """
+        @param full_sequence: on processor 0: the full sequence,
+                              equal to the concatenation of the local values
+                              of all processors. The local values
+                              on the other processors are not used.
+        @type full_sequence: L{ParValue}
+        """
+        messages = []
+        if processorID == 0:
+            full_sequence = full_sequence.value
+            length = len(full_sequence)
+            for pid in range(numberOfProcessors):
+                chunk = (length+numberOfProcessors-1)/numberOfProcessors
+                first = min(pid*chunk, length)
+                last = min(first+chunk, length)
+                value = full_sequence[first:last]
+                messages.append((pid, (first, last, length, value)))
+        send(messages)
+        self.first, self.last, self.length, self.value = retrieveMessages()[0]
+        self.valid = 1
+
 #
 # ParMessages serves to send exchange arbitray data between processors.
 #
 class ParMessages(ParValue):
 
-    """Global message list
-
-    A subclass of ParValue.
-
-    Constructor: ParMessage(|messages|)
-
-    Arguments:
-
-    |messages| -- a global object whose local value is a list of
-                  (pid, data) pairs.
+    """
+    Global message list
     """
 
     def __init__(self, messages):
+        """
+        @param messages: a global object whose local value is a list of
+                         (pid, data) pairs.
+        @type messages: Global sequence
+        """
         if hasattr(messages, 'is_parvalue'):
             messages = messages.value
         self.value = messages
         self.valid = 1
 
     def processorIds(self):
-        """Returns a ParValue object whose local value is a list of
-        all processor Ids referenced in a message.
+        """
+        @returns: a global object whose local value is a list of
+        all processor Ids referenced in a message
+        @rtype: L{ParValue}
         """
         return ParValue(map(lambda x: x[0], self.value))
 
     def data(self):
-        """Returns a ParValue object whose local value is a list of
+        """
+        @returns:  a global object whose local value is a list of
         all data items in the messages.
+        @rtype: L{ParValue}
         """
         return ParValue(map(lambda x: x[1], self.value))
 
     def exchange(self):
-        """Sends all the messages and returns a ParValue object
-        containing the received messages.
+        """
+        Transmit all the messages
+
+        @returns: a global object containing the received messages
+        @rtype: L{ParValue}
         """
         if self.valid:
             send(self.value)
-        return ParMessages(sync())
+        return ParMessages(retrieveMessages())
 
 #
 # ParTuple combines several ParValues to speed up communication.
 #
 class ParTuple(ParValue):
 
-    """Global data tuple
-
-    A subclass of ParValue.
-
-    Constructor: ParTuple(|x1|, |x2|, ...)
-
-    Arguments:
-
-    |x1|, |x2|, ... -- global objects
+    """
+    Global data tuple
 
     ParTuple objects are used to speed up communication when many data
     items need to be sent to the same processors. The construct
@@ -702,6 +831,10 @@ class ParTuple(ParValue):
     """
 
     def __init__(self, *args):
+        """
+        @param args: any global objects
+        @type args: C{tuple}
+        """
         self.value = map(lambda pv: pv.value, args)
         self.valid = reduce(operator.and_, map(lambda pv: pv.valid, args))
 
@@ -719,40 +852,39 @@ class ParTuple(ParValue):
 #
 class ParAccumulator(ParValue):
 
-    """Global accumulator
-
-    A subclass of ParValue.
-
-    Constructor: ParAccumulator(|operator|, |zero|)
-
-    Arguments:
-
-    |operator| -- a local function taking two arguments and returning
-                  one argument of the same type.
-
-    |zero| -- an initial value for reduction.
+    """
+    Global accumulator
 
     ParAccumulator objects are used to perform iterative reduction
-    operations in loops. The initial local value is |zero|, which is
+    operations in loops. The initial local value is zero (i.e. the
+    passed-in zero object, not the number 0), which is
     modified by subsequent calls to the method addValue.
     """
 
     def __init__(self, operator, zero):
+        """
+        @param operator: a local function taking two arguments and returning
+                         one argument of the same type
+        @type operator: function of two variables
+        @param zero: the initial value for reduction
+        """
         self.operator = operator
         self.zero = zero
         self.value = zero
         self.valid = 1
 
     def addValue(self, value):
-        """Replaces the internal value of the accumulator by
+        """
+        Replace the internal value of the accumulator by
         internal_value = operator(internal_value, value).
         """
         if value.valid:
             self.value = self.operator(self.value, value.value)
 
     def calculateTotal(self):
-        """Performs a reduction operation over the current local
-        values on all processors. Returns a ParValue object.
+        """
+        @returns: a reduction over the local values on all processors
+        @rtype: L{ParValue}
         """
         return self.reduce(self.operator, self.zero)
 
@@ -764,20 +896,17 @@ class ParFunction(ParValue):
 
     """Global function
 
-    A subclass of ParValue.
-
-    Constructor: ParFunction(|local_function|)
-
-    Arguments:
-
-    |local_function| -- a local function
-
     Global functions are called with global object arguments.
     The local values of these arguments are then passed to the local
     function, and the result is returned in a ParValue object.
     """
 
     def __init__(self, local_function):
+        """
+        @param local_function: any function
+        @type local_function: callable
+        """
+        local_function, is_valid = _getValue(local_function)
         self.value = local_function
         self.valid = 1
 
@@ -797,11 +926,6 @@ class ParRootFunction(ParFunction):
 
     Arguments:
 
-    |root_function| -- the local function for processor 0
-
-    |other_function| -- the local function for all other processors. The
-                        default is a function that returns None.
-
     Global functions are called with global object arguments.
     The local values of these arguments are then passed to the local
     function, and the result is returned in a ParValue object.
@@ -812,12 +936,20 @@ class ParRootFunction(ParFunction):
     """
 
     def __init__(self, local_function, other_function=None):
+        """
+        @param local_function: the local function for processor 0
+        @type local_function: callable
+        @param other_function: the local function for all other processors.
+                               The default is a function that returns None.
+        @type other_function: callable
+        """
+        local_function, is_valid = _getValue(local_function)
         if processorID == 0:
             self.value = local_function
         else:
             if other_function is None:
                 def other_function(*args, **kwargs):
-                    return ParValue(None)
+                    return RootNoneObject()
             self.value = other_function
         self.local_instance = None
         self.valid = 1
@@ -825,8 +957,14 @@ class ParRootFunction(ParFunction):
 #
 # ParIndex objects are returned by ParIndexIterator.
 #
-class ParIndex:
+class ParIndex(object):
 
+    """
+    Parallel index value
+
+    ParIndex objects are returned by ParIndexIterator. They are not meant
+    ot be created in any other way.
+    """
     def __init__(self, index, valid=1):
         if hasattr(index, 'is_parvalue'):
             self.valid = index.valid
@@ -850,6 +988,9 @@ class ParIndex:
 
 class ParSlice(ParIndex):
 
+    """
+    Parallel slice value
+    """
     def __init__(self, start=0, stop=None, skip=1, valid=1):
         self.start = start
         self.stop = stop
@@ -864,16 +1005,15 @@ class ParSlice(ParIndex):
 #
 # Direct iteration over distributed sequences.
 #
-class ParIterator:
+class ParIterator(object):
 
-    """Parallel iterator
+    """
+    Parallel iterator
 
     Constructor: ParIterator(|global_sequence|)
 
     Arguments:
 
-    |global_sequence| -- a global object representing a distributed
-                         sequence
 
     A ParIterator is used to loop element by element over a distributed
     sequence. At each iteration, the returned item (a global object)
@@ -881,6 +1021,10 @@ class ParIterator:
     """
 
     def __init__(self, sequence):
+        """
+        @param sequence: a global object representing a distributed sequence
+        @type sequence: L{ParSequence}
+        """
         self.sequence = sequence.value
         self.n = len(sequence.value)
         self.max_n = ParValue(self.n).reduce(max, 0).broadcast().value
@@ -895,25 +1039,23 @@ class ParIterator:
 #
 # Index iteration over distributed sequences.
 #
-class ParIndexIterator:
+class ParIndexIterator(object):
 
-    """Parallel index iterator
-
-    Constructor: ParIndexIterator(|global_sequence|)
-
-    Arguments:
-
-    |global_sequence| -- a global object representing a distributed
-                         sequence
+    """
+    Parallel index iterator
 
     A ParIndexIterator is used to loop index by index over one or more
     distributed sequences. At each iteration, the returned item (a
-    global index object) contains indices of different elements of the
+    L{ParIndex} object) contains indices of different elements of the
     distributed sequence(s). The index objects can be used to index
     any ParValue object whose local value is a sequence object.
     """
 
     def __init__(self, sequence):
+        """
+        @param sequence: a global object representing a distributed sequence
+        @type sequence: L{ParSequence}
+        """
         self.n = len(sequence.value)
         self.max_n = ParValue(self.n).reduce(max, 0).broadcast().value
 
@@ -928,20 +1070,15 @@ class ParIndexIterator:
 # Distribution class. This effectively turns all methods
 # into ParMethods and all other attributes inte ParValues.
 #
-class ParClass:
+class ParClass(object):
 
-    """Global class
-
-    Constructor: ParClass(|local_class|)
-
-    Arguments:
-
-    |local_class| -- a local class
+    """
+    Global class
 
     Global classes are needed to construct global objects that
     have more functionalities than offered by the ParValue class hierarchy.
     When an instance of a global class is generated, each processor
-    generates an instance of |local_class| that becomes the local value
+    generates an instance of the local class that becomes the local value
     of the new global object. Attribute requests and method calls
     are passed through to the local objects and the results are
     assembled into global objects (ParValue or ParFunction). The arguments
@@ -959,15 +1096,18 @@ class ParClass:
     """
 
     def __init__(self, local_class):
+        """
+        @param local_class: a standard Python class
+        """
         self.local_class = local_class
         self.attributes = {}
-        self.collectAttributes(local_class, self.attributes)
+        self._collectAttributes(local_class, self.attributes)
         try:
             del self.attributes['__init__']
         except KeyError:
             pass
 
-        class _Wrapper:
+        class _Wrapper(object):
             def __init__(self, local_instance):
                 self.value = local_instance
                 self.valid = 1
@@ -983,17 +1123,39 @@ class ParClass:
                                      repr(self.value))
             def __getattr__(self, name):
                 try:
-                    return ParValue(self.value.__dict__[name])
+                    return global_object(self.value.__dict__[name])
                 except KeyError:
                     pass
                 try:
                     value = self.attributes[name]
                 except KeyError:
                     value = getattr(self.value, name)
-                if type(value) == types.MethodType:
+                if isinstance(value, types.MethodType):
                     return ParMethod(value, self.value)
                 else:
-                    return ParValue(value)
+                    return global_object(value)
+            def __getitem__(self, item):
+                item, valid = _getValue(item)
+                if valid:
+                    return global_object(self.value[item])
+                else:
+                    return ParValue(None, 0)
+            def __call__(self, *args, **kwargs):
+                params = []
+                valid = True
+                for a in args:
+                    p, v = _getValue(a)
+                    valid = valid and v
+                    params.append(p)
+                kw = {}
+                for key, data in kwargs.items():
+                    p, v = _getValue(data)
+                    kw[key] = p
+                    valid = valid and v
+                if valid:
+                    return global_object(self.value(*params, **kw))
+                else:
+                    return ParValue(None, 0)
 
 
         self.wrapper = _Wrapper
@@ -1003,16 +1165,16 @@ class ParClass:
         self.wrapper.local_class = local_class
         _wrappers[local_class] = self.wrapper
 
-    def collectAttributes1(self, klass, attrib_dict):
+    def _collectAttributes1(self, klass, attrib_dict):
         for key in klass.__dict__.keys():
             if key not in ['__doc__', '__module__', '__name__', '__bases__']:
                 if not attrib_dict.has_key(key):
-                     attrib_dict[key] = getattr(klass, key)
+                    attrib_dict[key] = getattr(klass, key)
 
-    def collectAttributes(self, klass, attrib_dict):
-        self.collectAttributes1(klass, attrib_dict)
+    def _collectAttributes(self, klass, attrib_dict):
+        self._collectAttributes1(klass, attrib_dict)
         for base_class in klass.__bases__:
-            self.collectAttributes(base_class, attrib_dict)
+            self._collectAttributes(base_class, attrib_dict)
             
     def __call__(self, *args, **kwargs):
         args = (processorID, numberOfProcessors) + args
@@ -1022,21 +1184,18 @@ class ParClass:
         return self.wrapper(local_instance)
 
 # Dummy class
-class _DummyClass:
+class _DummyClass(object):
     pass
 
 #
 # Special 'invalid' object. It is passed to methods in distributed
 # object classes and accepted as a return value.
 #
-class _ParInvalid:
+class _ParInvalid(object):
     pass
 
-def is_invalid(object):
-    try:
-        return object.__class__ is _ParInvalid
-    except AttributeError:
-        return 0
+def is_invalid(obj):
+    return isinstance(obj, _ParInvalid)
 
 ParInvalid = _ParInvalid()
 _wrappers[_ParInvalid] = lambda x: ParValue(None, 0)
@@ -1047,6 +1206,13 @@ _wrappers[_ParInvalid] = lambda x: ParValue(None, 0)
 #
 class ParMethod(ParFunction):
 
+    """
+    Method of a global class
+
+    ParMethod objects are created by ParClass. They are not meant to be used
+    directly in application code.
+    """
+
     def __init__(self, local_function, local_instance):
         self.value = local_function
         self.local_instance = local_instance
@@ -1080,9 +1246,10 @@ class ParMethod(ParFunction):
 #
 # Abstract base class that provides communication for ParClasses.
 #
-class ParBase:
+class ParBase(object):
 
-    """Distributed data base class
+    """
+    Distributed data base class
 
     Local classes that are to be used in global classes
     must inherit from this class.
@@ -1091,33 +1258,58 @@ class ParBase:
     is_parclass = 1
 
     def put(self, data, pid_list):
-        """Send |data| to all processors in |pid_list|. Returns the list
-        of received objects.
+        """
+        Send data to other processors
+
+        @param data: the data to be sent
+        @type data: any
+        @param pid_list: the list of processor numbers to which the data is sent
+        @type pid_list: C{list}
+        @returns: the values received from other processors
+        @rtype: C{list}
         """
         put(data, pid_list)
-        return sync()
+        return retrieveMessages()
 
     def get(self, data, pid_list):
-        """Requests the local values of |data| of all processors in |pid_list|.
-        Returns the list of received objects.
+        """
+        Request the local values of other processors.
+
+        @param data: the data to be sent to processors who request it
+        @type data: any
+        @param pid_list: the list of processor numbers to which data requests
+                         are sent
+        @type pid_list: C{list}
+        @returns: the values received from other processors
+        @rtype: C{list}
         """
         put(processorID, pid_list)
         destinations = sync()
         put(data, destinations)
-        return sync()
+        return retrieveMessages()
 
     def broadcast(self, data, from_pid=0):
-        """Sends the local value of |data| on processor |from_pid| to
-        all processors. Returns the list of received objects.
+        """
+        Send a local value of one processor to all processors.
+
+        @param data: the data to be transmitted. This value is used
+                     only on one processor.
+        @param from_pid: the processor whose data is broadcast
+        @type from_pid: any
+        @returns: the received data
+        @rtype: any
         """
         if processorID == from_pid:
             put(data, range(numberOfProcessors))
-        return sync()[0]
+        return retrieveMessages()[0]
 
     def exchangeMessages(self, message_list):
-        """Sends the (pid, data) messages in |message_list| to the destination
-        processors. Returns the list of incoming data.
+        """
+        @param message_list: a list of (pid, data) pairs to be transmitted
+        @type message_list: C{list}
+        @returns: the incoming data
+        @rtype: C{list}
         """
         send(message_list)
-        return sync()
+        return retrieveMessages()
 
